# Copyright 2023 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""Natively supported multimodal datasets."""

from streaming.multimodal.webvid import StreamingWebVid as StreamingWebVid
